
import warnings
import numpy as np
import pandas
from . import io
from astropy import time
        
def get_sample(**kwargs):
    """ Short to to Sample.load() """
    return Sample.load(**kwargs)


class Sample():
    
    def __init__(self, data=None):
        """ """
        self.set_data(data)
    
    @classmethod
    def load(cls, redshift_range=None, target_list=None):
        """ Load a Sample instance building it from io.get_targets_data() """
        data = io.get_targets_data()
        if redshift_range is not None:
            data = data[data["redshift"].between(*redshift_range)]

        if target_list is not None:
            data = data.loc[target_list]
            
        return cls(data=data)

    # ------- #
    # LOADER  #
    # ------- #
    def load_hostdata(self):
        """ load the host data using io.get_host_data(). This is made automatically upon hostdata call. """
        self.set_hostdata( io.get_host_data() )

    def load_spectra_df(self, add_phase=True, current_target=True):
        """ load the spectra dataframe 

        Parameters
        ----------
        add_phase: bool
            should the phase (based on spectra date and the target t0) be added

        current_target: bool
            only keep target that are in the current sample.
        
        Returns
        -------
        None
            sets self.spectra_df
        """
        spectra_df = io.get_spectra_datafile().set_index("ztfname")
        if current_target:
            spectra_df = spectra_df.loc[spectra_df.index.isin( self.data.index )]
            
        spectra_df["mjd"] = time.Time(np.asarray(spectra_df["date"].apply(lambda x: f"{x[:4]}-{x[4:6]}-{x[6:]}").astype(str).values, dtype="str")).mjd
        if add_phase:
            spectra_df = spectra_df.join( self.data["t0"] )
            spectra_df["phase"] = spectra_df["mjd"] - spectra_df["t0"]

        self._spectra_df = spectra_df
        
    def load_phase_df(self, min_detection=5, groupby='filter', client=None, rebuild=False, **kwargs):
        """ Load the phase_df. This is made automatically upton phase_df call.
        If this is the first time ever you call this, the phase_df has to be built, so it takes ~45s.
        Once built, the dataframe is stored such that, next time, it is directly loaded. 
        Use rebuild=True to bypass the automatic loading and rebuild the dataframe.
        
        Parameters
        ----------
        min_detection: [float] -optional-
            minimal signal to noise ratio for a point to be considered as 'detected'

        groupby: [string or list of] -optional-
            data column(s) to group the lightcurve data and measuring the count statistics.
            example: filter, [filter, fieldid]
             
        client: [dask.Client] -optional-
            if a dask.distributed.Client instance is given, this will use dask. 
            Otherwise a basic for loop (slower) is used.

        rebuild: [bool] -optional-
            force the dataframe to be rebuilt.

        **kwargs goes to self.build_phase_coverage()

        Returns
        -------
        None
        """
        if not rebuild:
            phase_df = io.get_phase_coverage(load=True, warn=False)
            if phase_df is None:
                rebuild = True
                
        if rebuild:
            phase_df = self.build_phase_coverage(min_detection=min_detection,
                                                    groupby=groupby,
                                                    client=client, store=True, **kwargs)
        self._phase_df = phase_df

    def load_fiedid(self):
        """ compute the fields containing the targets using 
        ztfquery.fields.get_fields_containing_target().
        This take quite some time.

        the 'fieldid' entry is added to self.data
        """
        from ztfquery import fields
        fieldid = [fields.get_fields_containing_target(s_["ra"],s_["dec"])
                       for i_,s_ in self.data.iterrows()]
        self.data["fieldid"] = fieldid

    # ------- #
    # SETTER  #
    # ------- #
    def set_data(self, data):
        """ attach to this instance the target data. 
        = Most likely you should not use this method directly. =
        use sample = Sample.load()
        """
        data["t0day"] = data["t0"].astype("int")
        self._data = data

    def set_hostdata(self, hostdata):
        """ attach to this instance the host data.
        = Most likely you should not use this method directly. =
        use sample.load_hostdata() or simply 
        sample.hostdata that automatically load this corrent hostdata.
        """
        self._hostdata = hostdata
        
    def merge_to_data(self, dataframe, how="outer", **kwargs):
        """ Merge the given dataframe with self.data. 
        The merged dataframe will replace self.data
        """
        self._data = pandas.merge(self.data, dataframe, how=how, 
                                 **{**dict(left_index=True, right_index=True),
                                 **kwargs})
        
    # ------- #
    # GETTER  #
    # ------- #
    def get_data(self, clean_t0nan=True,
                     t0_range=None, x1_range=None, c_range=None,
                     redshift_range=None, z_quality=None, redshift_source=None,
                     t0_err_range=None, c_err_range=None, x1_err_range=None,
                     exclude_targets=None, in_targetlist=None,
                     ndetections=None,
                     goodcoverage=None, coverage_prop={},
                     first_spec_phase=None, query=None, data=None):
        """ 
        *_range: [None or [min, max]] -optional-
            cut to be applied to the data for
            t0, x1, c, (and there error) and redshift.
            boundaries are mandatory. For instance, redshift range lower than 0.06
            should be: redshift_range = (0, 0.06).
            = see below for t0 format =
        
        t0_range: [None or [tmin,tmax]] -optional-
            Should be a format automatically understood by astropy.time.Time
            e.g. t0_range=["2018-04-01","2020-10-01"]

        in_targetlist: [list of strings] -optional-
            The target must be in this list

        goodcoverage: [None or bool] -optional-
            Select the data given the lc phase coverage
            - None: no cut
            - True: only good lc kept
            - False: only bad lc kept (good lc discarded)
            This uses self.get_goodcoverage_targets(**coverage_prop)

        coverage_prop: [dict] -optional-
            kwargs passed to self.get_goodcoverage_targets
            = used only if goodcoverage is not None =

        query: [string] -optional-
            any additional query to be given to data.query({query}).
            This are SQL-like format applied to the colums. 
            See pandas.DataFrame.query()


        Returns
        -------
        DataFrame (sub part or copy of self.data)
        """
            
        if clean_t0nan:
            data = self.data[~self.data["t0"].isna()]
        else:
            data = self.data.copy()
        
        # - Time Range
        if t0_range is not None:
            t0_start = time.Time(t0_range[0]).mjd
            t0_end = time.Time(t0_range[1]).mjd
            data = data[data["t0"].between(t0_start, t0_end)]

        # LC Cuts
        # - stretch (x1) range
        if x1_range is not None:
            data = data[data["x1"].between(*x1_range)]

        # - color (c) range
        if c_range is not None:
            data = data[data["c"].between(*c_range)]

        # - t0 errors  range
        if t0_err_range is not None:
            data = data[data["t0_err"].between(*t0_err_range)]
            
        # - stretch errors (x1) range
        if x1_err_range is not None:
            data = data[data["x1_err"].between(*x1_err_range)]

        # - color errors (c) range
        if c_err_range is not None:
            data = data[data["c_err"].between(*c_err_range)]


        # Redshift Cuts
        # - Redshift range
        if redshift_range is not None:
            data = data[data["redshift"].between(*redshift_range)]
            
        # -  redshift origin
        if redshift_source is not None:
            data = data[data["redshift"].isin(*np.atleast_1d(redshift_source))]
            
        if z_quality is not None and z_quality not in ["any","all","*"]:
            data = data[data["z_quality"].isin(np.atleast_1d(z_quality))]

        # Target cuts
        # - in given list
        if in_targetlist is not None:
            data = data.loc[np.asarray(in_targetlist)[np.in1d(in_targetlist, data.index.astype("string"))] ]

        if exclude_targets is not None:
            data = data.loc[~data.index.isin(exclude_targets)]


        # LC Coverage
        if ndetections is not None:
            phase_coverage = self.get_phase_coverage()["n_points"]
            min_corevarge = phase_coverage[phase_coverage>=ndetections]
            data = data.loc[data.index.isin(min_corevarge.index)]
            
        # - special good lc list.            
        if goodcoverage is not None:
            good_covarege_targets = self.get_goodcoverage_targets(**coverage_prop)
            # doing it with np.in1d to make sure all matches and not some are already missing
            flag_goodcoverage = np.asarray(good_covarege_targets)[np.in1d(good_covarege_targets, data.index.astype("string"))]
            if goodcoverage:
                data = data.loc[flag_goodcoverage]
            else:
                data = data.loc[~flag_goodcoverage]

        # Spectral Cut.
        if first_spec_phase is not None:
            first_spec = self.spectra_df.groupby(level=0).phase.first()
            first_spec = first_spec[first_spec<=first_spec_phase]
            data = data.loc[data.index.isin(first_spec.index)]

        # Additional Query
        if query:
            data = data.query(query)
            
        return data

    # LightCurve
    def get_target_lightcurve(self, name):
        """ Get the {name} LightCurve object """
        from . import lightcurve
        return lightcurve.LightCurve.from_name(name)

    # Spectrum
    def get_target_spectra(self, name):
        """ Get a list with all the Spectra for the given object """
        from . import spectroscopy
        return spectroscopy.Spectrum.from_name(name)

    # Extra
    def get_goodcoverage_targets(self,
                                     n_early_bands=">=2",
                                     n_late_bands=">=2",
                                       n_points=">=7",
                                       premax_range=[-15,0],
                                       postmax_range=[0,30],
                                       phase_range=[-15,30],
                                       **kwargs):
        """ kwargs should have the same format as the n_early_point='>=2' for instance.
        None means no constrain, like n_bands=None means 'n_bands' is not considered.
        """
        query = {**dict(n_early_bands=n_early_bands, n_late_bands=n_late_bands,
                        n_points=n_points),
                 **kwargs}
        df_query = " and ".join([f"{k}{v}" for k,v in query.items() if v is not None])
        phase_coverage = self.get_phase_coverage(premax_range=premax_range,
                                                 postmax_range=postmax_range,
                                                 phase_range=phase_range)
        return phase_coverage.query(df_query).index.astype("string")

    
    def get_phase_coverage(self,premax_range=[-15,0],
                                postmax_range=[0,30],
                                phase_range=[-15,30], min_det_perband=1):
        """ """        
        # All
        phases = self.phase_df[self.phase_df.between(*phase_range)].reset_index().rename({"level_0":"name"},axis=1)
        n_points = phases.groupby(["name"]).size().to_frame("n_points")
        n_bands = (phases.groupby(["name", "filter"]).size()>=min_det_perband
                        ).groupby(level=[0]).sum().to_frame("n_bands")
        # Pre-Max
        # - AnyBand
        premax = self.phase_df[self.phase_df.between(*premax_range)].reset_index().rename({"level_0":"name"},axis=1)
        n_early_points = premax.groupby(["name"]).size().to_frame("n_early_points")
        n_early_bands = (premax.groupby(["name", "filter"]).size()>=min_det_perband
                        ).groupby(level=[0]).sum().to_frame("n_early_bands")
        # - Per filters
        n_early_points_perfilter = premax.groupby(["name", "filter"]).size()
        n_early_points_g = n_early_points_perfilter.xs("p48g", level=1).to_frame("n_early_points_p48g")
        n_early_points_r = n_early_points_perfilter.xs("p48r", level=1).to_frame("n_early_points_p48r")
        n_early_points_i = n_early_points_perfilter.xs("p48i", level=1).to_frame("n_early_points_p48i")
        
        # Post-Max
        # - AnyBand        
        postmax = self.phase_df[self.phase_df.between(*postmax_range)].reset_index().rename({"level_0":"name"},axis=1)
        n_late_points = postmax.groupby(["name"]).size().to_frame("n_late_points")
        n_late_bands = (postmax.groupby(["name", "filter"]).size()>=min_det_perband
                       ).groupby(level=[0]).sum().to_frame("n_late_bands")
        # - Per filters
        n_late_points_perfilter = postmax.groupby(["name", "filter"]).size()
        n_late_points_g = n_late_points_perfilter.xs("p48g", level=1).to_frame("n_late_points_p48g")
        n_late_points_r = n_late_points_perfilter.xs("p48r", level=1).to_frame("n_late_points_p48r")
        n_late_points_i = n_late_points_perfilter.xs("p48i", level=1).to_frame("n_late_points_p48i")
            
        return pandas.concat([n_points,n_bands,
                              n_early_points,n_early_bands,n_late_points, n_late_bands,
                              n_early_points_g, n_early_points_r, n_early_points_i,
                              n_late_points_g, n_late_points_r, n_late_points_i], axis=1).fillna(0).astype(int)

        
    def build_phase_coverage(self, min_detection=5, groupby='filter', client=None, store=True, **kwargs):
        """ 
        time: 
        - Dask: 45s on a normal laptop | 4 cores 
        - No Dask: 60s on a normal laptop | 4 cores 
        """
        import pandas
        import dask
        from . import lightcurve

        phases = []
        datalc = self.get_data()
        datalc = datalc[datalc["redshift"].between(-0.1,0.2)]

        warnings.warn("building phase coverage takes ~30s to 1min.")
        # - Without Dask
        if client is None:
            warnings.warn("loading without Dask (client is None) ; it will be slow")
            names_ok = []
            for name in datalc.index:
                try:
                    dt = lightcurve.LightCurve.from_name(name)
                    phases.append( dt.get_obsphase(min_detection=min_detection, groupby=groupby, **kwargs))
                    names_ok.append(name)
                except:
                    warnings.warn(f"get_obsphase did not work for {name}")
                    continue
                
            phase_df = pandas.concat(phases, keys=names_ok)

        # - With Dask
        else:
            for name in datalc.index:
                dt = dask.delayed(lightcurve.LightCurve.from_name)(name)
                phases.append( dt.get_obsphase(min_detection=min_detection, groupby=groupby, **kwargs)
                             )

            fphases = client.compute(phases)
            data_ = client.gather(fphases, "skip") # wait until all is done
            names = datalc.index[[i for i,f_ in enumerate(fphases) if f_.status=="finished"]]
            
            phase_df = pandas.concat(data_, keys=names)

        phase_df_exploded = phase_df.explode()
        if store:
            filepath = io.get_phase_coverage(load=False)
            phase_df_exploded.to_csv(filepath)
            
        return phase_df_exploded
    
    
    # ------- #
    # PLOTTER #
    # ------- #
    def show_ideogram(self, key, keyerr=None,
                          ax=None, data=None, dataprop={},
                          facecolor=None, edgecolor=None, lw=2,
                          nsigma_range=3, npoints=1000,
                          density=False, normed_one=False,
                          **kwargs):
        """ """
        from matplotlib.colors import to_rgba
        from scipy import stats

        # axes
        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[6,3])
            ax = fig.add_axes([0.1,0.2,0.8,0.7])
        else:
            fig = ax.figure

        # data            
        if keyerr is None:
            keyerr = key+"_err"

        if data is None:
            data = self.get_data(**dataprop)

        # getting the x and y            
        d, d_err = data[[key, keyerr]].values.T
        d_range = np.nanmin(d-nsigma_range*d_err), np.nanmax(d+nsigma_range*d_err)
        dd = np.linspace(*d_range, npoints)
        dideo = np.sum(stats.norm.pdf(dd[:,None],
                                      loc=d, scale=d_err), 
                           axis=1)
        
        # plotting properties
        if facecolor is None:
            facecolor = to_rgba("C0",0.3)
            facecolor_ = "C0"
        else:
            facecolor_ = facecolor
            
        if edgecolor is None:
            edgecolor = to_rgba(facecolor_, 1)


        if density:
            dideo /= np.nansum(dideo)
            
        if normed_one:
            dideo /= np.nanmax(dideo)
            
        # plot
        _ = ax.fill_between(dd, dideo, 
                                facecolor=facecolor, 
                                edgecolor=edgecolor, lw=lw,
                                **kwargs)
        
        # fancy
        ax.set_ylim(0)
        return fig
    
    def show_discoveryhist(self, ax=None, daymax=15, linecolor="C1", **kwargs):
        """ """
        from matplotlib.colors import to_rgba
        datasalt = self.get_data(clean_t0nan=True)
        
        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[6,3])
            ax = fig.add_axes([0.1,0.2,0.8,0.7])
        else:
            fig = ax.figure
            

        prop = dict(fill=True, histtype="bar", 
                    density=False, facecolor=to_rgba("C0", 0.1), edgecolor="C0",
                   align="left", zorder=3)
        _ = ax.hist(datasalt.groupby("t0day").size(), range=[0,15], bins=15,
                    **{**prop,**kwargs})
        if linecolor != "None":
            ax.axvline(3.4, color="C1", zorder=5, lw=1.5, ls="--")
        
#        xx = np.arange(0, daymax)        
        #ax.scatter(xx[1:], stats.poisson.pmf(xx[1:], mu=3.5)*1050, 
        #           s=50, color="C1", zorder=5)

        #ax.set_yscale("log")

        _ = ax.set_xticks(np.arange(daymax))
        ax.set_xlim(0)

        clearwhich = ["left","right","top"] # "bottom"
        [ax.spines[which].set_visible(False) for which in clearwhich]
        ax.tick_params("y", labelsize="x-small")
        ax.set_yticks([0,100,200])

        ax.set_xlabel("Number of SN Ia per day")
        return fig
    
    
    def show_discoveryevol(self, ax=None, t0_range=["2018-04-01","2021-01-01"],
                           typed_color="C0", quality_color="goldenrod",
                           xformat=True, dataprop={}, 
                           **kwargs):
        """ """
        from matplotlib.colors import to_rgba

        # - Data    
        datasalt_all = self.get_data(clean_t0nan=True, 
                                        t0_range=t0_range,
                                        **dataprop)
        
        datasalt_lccut = self.get_data(clean_t0nan=True,
                                        t0_range=t0_range,
                                        goodcoverage=True,
                                        **dataprop)
        
        datasalt_zcut = self.get_data(clean_t0nan=True,
                                        t0_range=t0_range, 
                                        z_quality=2,
                                        **dataprop)
        
        datasalt_zcut_lccut = self.get_data(clean_t0nan=True,
                                        t0_range=t0_range, 
                                        z_quality=2,
                                        goodcoverage=True,
                                        **dataprop)

        
        # - Figue
        if ax is None:
            import matplotlib.pyplot as mpl            
            fig = mpl.figure(figsize=[8,4])
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure


        # - Internal

        def _show_single_(gb_cumsize, mplfunc="fill_between", 
                          add_text=True, text_color=None, **kwargs_):
            """ """
            time_ = time.Time(gb_cumsize.index, format="mjd").datetime
            values_ = gb_cumsize.values
            _ = getattr(ax,mplfunc)(time_, values_, **kwargs_)
            if add_text:
                ax.text(time_[-1], values_[-1],f" {values_[-1]}",
                            va="center",ha="left", color=text_color)

        # Show all
        _show_single_(datasalt_all.groupby("t0day").size().cumsum(), 
                     lw=2, ls="None", facecolor="w", edgecolor="None", 
                      add_text=False)
        _show_single_(datasalt_all.groupby("t0day").size().cumsum(), 
                     lw=2, ls="--", color=typed_color, mplfunc="plot",
                     text_color=typed_color)



        _show_single_(datasalt_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="None", facecolor=to_rgba(quality_color,0.2), 
                      edgecolor="None", add_text=False)
        _show_single_(datasalt_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="--", color=quality_color, mplfunc="plot",
                     text_color=quality_color)



        _show_single_(datasalt_zcut.groupby("t0day").size().cumsum(), 
                      mplfunc="plot",
                     lw=1, ls="-", color=typed_color, 
                     text_color=typed_color)

        _show_single_(datasalt_zcut_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="-", facecolor=to_rgba(quality_color,0.2), 
                      edgecolor="None", add_text=False)
        _show_single_(datasalt_zcut_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="-", color=quality_color, mplfunc="plot",
                     text_color=quality_color)


        # - Out Formating
        if xformat:
            from matplotlib import dates as mdates
            locator = mdates.AutoDateLocator()
            formatter = mdates.ConciseDateFormatter(locator)
            ax.xaxis.set_major_locator(locator)
            ax.xaxis.set_major_formatter(formatter)

        clearwhich = ["left","right","top"] # "bottom"
        [ax.spines[which].set_visible(False) for which in clearwhich]

    #        ax.set_title("ZTF-1 Ia Statistics")
        ax.set_ylabel("Number of Type Ia Supernovae")
        ax.set_ylim(bottom=0)
        ax.tick_params("y", labelsize="small")
        return fig
        
    # =============== #
    #   Properties    #
    # =============== #
    @property
    def data(self):
        """ """
        return self._data

    @property
    def hostdata(self):
        """ """
        if not hasattr(self, "_hostdata"):
            self.load_hostdata()
        return self._hostdata


    @property
    def phase_df(self):
        """ """
        if not hasattr(self, "_phase_df"):
            self.load_phase_df()
        
        return self._phase_df

    @property
    def spectra_df(self):
        """ dataframe containing the spectral information """
        if not hasattr(self, "_spectra_df"):
            self.load_spectra_df()
        
        return self._spectra_df




# Loads the ztf filters inside sncosmo
import warnings
from ztfquery import filters
import numpy as np
import pandas
filters.load_p48_filters_to_sncosmo(basename="p48") # p48g etc.


from .io import get_salt2params
try:
    SALT2PARAMS = get_salt2params()
except:
    warnings.warn("Failed to load SALT2 parameters")
    SALT2PARAMS = None



def salt2result_to_dataframe(result):
    """ """
    fitted = np.in1d(result.param_names, result.vparam_names)
    df = pandas.DataFrame(np.asarray([result.parameters, fitted]).T, 
                          result.param_names, 
                          columns=["values", "fitted"])
    df.fitted = df.fitted.astype(bool)
    
    # - Error
    df = df.merge(pandas.Series(dict(result["errors"]), name="errors"), 
                      left_index=True, right_index=True, how="outer")
    
    # - Cov
    dcov = pandas.DataFrame(result["covariance"], columns=result.vparam_names, index=result.vparam_names)
    dcov.columns ="cov_"+dcov.columns
    
    # - merged
    return df.merge(dcov,  left_index=True, right_index=True, how="outer")
    
# ================== #
#                    #
#   PARAMETERS       #
#                    #
# ================== #
def get_target_salt2param(targetname):
    """ """
    if SALT2PARAMS is not None:
        if targetname in SALT2PARAMS.index:
            return SALT2PARAMS.loc[targetname]
        else:
            warnings.warn(f"{targetname} is not in the SALTPARAMS dataframe")

    return None

# ================== #
#                    #
#   MODEL            #
#                    #
# ================== #
def get_saltmodel(**params):
    """ """
    import sncosmo
    dust  = sncosmo.CCM89Dust()
    model = sncosmo.Model("salt2", effects=[dust],
                              effect_names=['mw'],
                              effect_frames=['obs'])
    model.set(**params)
    return model



import numpy as np
import os
import pandas
import warnings
from . import io

SPEC_DATAFILE = io.get_spectra_datafile()
TARGETS_DATA = io.get_targets_data()

__all__ = ["get_target_spectra"]

def get_target_spectra(name, load_snidres=True, **kwargs):
    """ load the lightcurve object for the given target. """
    return np.atleast_1d(Spectrum.from_name(name, load_snidres=load_snidres,
                                            **kwargs)
                        )



def read_spectrum(file_, sep=None):
    """ """
    data = open(file_).read().splitlines()
    try:
        header = pandas.DataFrame([d.replace("#","").replace(":"," ").replace("=","").split()[:2] for d in data 
                                   if (d.startswith("#") or "=" in d or ":" in d) and len(d.split())>=2],
                              columns=["keys", "values"]).set_index("keys")["values"]
    except:
        raise IOError(f"header build Fails for {file_}")
    
    try:
        lbda, flux, *variance = np.asarray([d.split(sep) for d in data 
                                        if not (d.startswith("#") 
                                                or d.startswith("COMMENT") or d.startswith("HISTORY")
                                                or "=" in d or ":" in d) and len(d.split())>=2]).T
        data = pandas.DataFrame({"lbda":lbda, "flux":flux}, dtype="float")
    except:
        raise IOError(f"dataframe build Fails for {file_}")
    if len(variance)>0:
        data["variance"] = variance[0]
    return header, data




class Spectrum( object ):
    """ """
    def __init__(self, data, header=None, meta=None, use_dask=False, snidresult=None,
                     filename=None, load_snidres=True):
        """ """
        self.set_data(data)
        self.set_header(header)
        self.set_meta(meta)
        self.set_snidresult(snidresult)
        # - Dask
        self._use_dask = use_dask
        # - Filename        
        self._filename = filename
        # - snid result
        if load_snidres:
            self.load_snidresult(allow_fit=False)

        
    @classmethod
    def from_filename(cls, filename, use_dask=False, snidresult=None,
                          load_snidres=True, sep=None, **kwargs):
        """ 
        load_snidres: fetch for snidresults and loads it if it exists.
        """
        if not use_dask:
            header, data = read_spectrum(filename, sep=sep)
        else:
            from dask import delayed
            header_data = delayed(read_spectrum)(filename)
            header = header_data[0]
            data = header_data[1]
            
        meta = io.parse_filename(filename)
        return cls(data, header=header, meta=meta, use_dask=use_dask,
                   snidresult=snidresult, filename=filename,
                   load_snidres=load_snidres, **kwargs)

    @classmethod
    def from_name(cls, targetname, as_spectra=False, use_dask=False,
                      load_snidres=True, **kwargs):
        """ """        
        fullpath = SPEC_DATAFILE[SPEC_DATAFILE["ztfname"]==targetname]["fullpath"].values
        if len(fullpath)==0:
            warnings.warn(f"No spectra target names {targetname}")
            return None
        
        if len(fullpath) == 1:
            return cls.from_filename(fullpath[0], use_dask=use_dask, load_snidres=load_snidres,
                                         **kwargs)
        elif as_spectra:
            return Spectra.from_filenames(fullpath, use_dask=use_dask, load_snidres=load_snidres,
                                         **kwargs)
        else:
            return [cls.from_filename(file_, use_dask=use_dask, load_snidres=load_snidres,
                                         **kwargs)
                        for file_ in fullpath]


    def store(self, fileout=None, overwrite=False):
        """ """
        if os.path.isfile(fileout) and not overwrite:
            raise IOError("fileout already exists. Set overwrite=True to overwrite it.")

        nlbda = len(self.lbda)
        with open(fileout,"w") as file:
            if self.header is not None:
                for i_,s_ in self.header.iteritems():
                    file.write(f"# {i_}: {s_}\n")
                
            if self.has_variance():
                for i_,(w_,f_,v_) in enumerate(zip(self.lbda, self.flux, self.variance)):
                    file.write(f"{w_} {f_} {v_}")
                    if not i_ == nlbda-1:
                        file.write("\n")
                        
            else:
                for i_,(w_,f_) in enumerate(zip(self.lbda, self.flux)):
                    file.write(f"{w_} {f_}")
                    if not i_ == nlbda-1:
                        file.write("\n")

                    
    # ================ #
    #    Method        #
    # ================ #
    def fetch_snidresult(self, warn_if_notexist=True):
        """ """
        if self.filename is None:
            raise AttributeError("Unknown filename for the given spectrum. Cannot fetch the corresponding snidresult")

        snidresult_file = self.filename.replace(".ascii","_snid.h5")
        if not os.path.isfile(snidresult_file):
            if warn_if_notexist:
                warnings.warn(f"snidres file does not exists {snidresult_file}")
            return None
        
        from pysnid.snid import SNIDReader
        return SNIDReader.from_filename(snidresult_file)
    
    def load_snidresult(self, phase=None, redshift=None,
                            force_fit=False, allow_fit=True, **kwargs):
        """ get and set """
        if force_fit:
            allow_fit = True
            
        snidresult = self.fetch_snidresult(warn_if_notexist=False)
        if (snidresult is None or force_fit) and allow_fit:
            snidresult = self.get_snidfit(phase=phase, redshift=redshift,
                                       **kwargs)
        self.set_snidresult( snidresult )
        
    # --------- #
    #  SETTER   #
    # --------- #
    def set_data(self, data):
        """ """
        self._data = data
        
    def set_header(self, header):
        """ """
        self._header = header
    
    def set_meta(self, meta):
        """ """
        self._meta = meta

    def set_snidresult(self, snidresult):
        """ """
        self._snidresult = snidresult
        
    # --------- #
    #  GETTER   #
    # --------- #
    def get_obsdate(self):
        """ calls self.obsdate """
        return self.obsdate
    
    def get_phase(self, t0=None, z=None, from_target=True):
        """ t0, the phase=0 date, is expected to be in MJD. 
        This returns phase=obsdate - t0
        If z, the redshift is given, then phase/=(1+z) to have 
        it in rest-frame.

        if from_target, t0 and z are loaded from io.get_target_data()
        """
        # = Where does t0 come from ?
        if from_target:
            if self.targetname is None:
                raise AttributeError("The targetname is unknown, self.targetname is None. You must provide t0 (and z) manually.")
            t0, z = TARGETS_DATA.loc[self.targetname][["t0","redshift"]].values
                
        elif t0 is None:
            raise ValueError("from_target is False and no t0 given.")

        # = Ok, let's go.        
        obsdate = self.get_obsdate()
        if obsdate is None:
            return None
        
        phase = obsdate.mjd-t0
        if z is not None:
            phase /=(1+z)
            
        return phase

    def get_snidfit(self, phase=None, redshift=None,
                        delta_phase=5, delta_redshift=None,
                        lbda_range=[4000, 8000], **kwargs):
        """ """
        warnings.warn("DEPRECATED: use self.fit_snid() instead. This function redirect to it now. ")
        return self.fit_snid(phase=phase, redshift=redshift,
                                 delta_phase=delta_phase, delta_redshift=delta_redshift,
                                 lbda_range=lbda_range, **kwargs)

    def get_redshift(self):
        """ """
        if self.snidresult is None:
            warnings.warn("snidres is not defined (None)")
            return [np.nan, np.nan]

        return self.snidresult.get_redshift()

    
    # --------- #
    #  FITTER   #
    # --------- #
    def fit_snid(self,  phase=None, redshift=None,
                        delta_phase=5, delta_redshift=None,
                        lbda_range=[4000, 8000], **kwargs):
        """ """
        import pysnid
        if self.filename is None:
            raise NotImplementedError("No filename stored (self.filename is None). No work around implemented")
            
        return pysnid.run_snid(self.filename, redshift=redshift, phase=phase,
                                   delta_phase=delta_phase, delta_redshift=delta_redshift,
                                   lbda_range=lbda_range, **kwargs)
    
    # --------- #
    # PLOTTER   #
    # --------- #        
    def show(self, ax=None, savefile=None, color=None, ecolor=None, ealpha=0.2, 
             show_error=True, zeroline=True, zcolor="0.7", zls="--", 
             zprop={}, fillprop={}, normed=False, offset=None,
             label=None, legend=None, **kwargs):
        """ 
        label: [string or None]
            label for the spectra. 
            use: label='_meta_' to see meta content.
            
        """

        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[6,4])
            ax = fig.add_axes([0.12,0.15,0.78,0.75])
        else:
            fig = ax.figure
        
        self._compute()
        prop = dict(zorder=3)
        coef = 1 if not normed else self.flux.mean()
        if offset is None:
            offset = 0

        if label == "_meta_":
            label = f"{self.meta['name']} | by {self.meta.instrument} on {self.meta.date.split('T')[0]}"
        elif label == "_meta_noname_":
            label = f"{self.meta.instrument} on {self.meta.date.split('T')[0]}"
            
        if label is not None and legend is None:
            legend=True
            
        _ = ax.plot(self.lbda, self.flux/coef + offset, label=label, color=color, **{**prop, **kwargs})
        if self.has_error() and show_error:
            if ecolor is None:
                ecolor = color
            ax.fill_between(self.lbda, (self.flux-self.error)/coef+ offset, (self.flux+self.error)/coef+ offset, 
                           facecolor=ecolor, alpha=ealpha, **{**prop,**fillprop})
            
        if zeroline:
            ax.axhline(0, color=zcolor,ls=zls, **{**dict(lw=1, zorder=1),**zprop} )
            
        ax.set_ylabel("Flux []"+ " -normed-" if normed else "")
        ax.set_xlabel(r"Wavelength [$\AA$]")
        if legend:
            ax.legend(loc="best", frameon=False, fontsize="small")

        if savefile:
            fig.savefig(savefile, dpi=150)
            
        return fig

    def show_snidresult(self, axes=None, savefile=None, label=None, **kwargs):
        """ shortcut to self.snidresult.show() """
        if self.snidresult is None:
            warnings.warn("snidres is not defined (None)")
            return self.show(ax=axes[0])
        
        return self.snidresult.show(axes=axes, savefile=savefile, label=label, **kwargs)
    
    # ------------ #
    #   Internal   #
    # ------------ #
    def _compute(self):
        """ """
        if not self._use_dask:
            return
        from dask.delayed import Delayed
        from dask import delayed
        if not type(self._data) == Delayed:
            return 
        self._data, self._header = delayed(list)([self._data, self._header]).compute()
        
    # ================ #
    #   Properties     #
    # ================ #
    # Baseline    
    @property
    def data(self):
        """ """
        return self._data
    
    @property
    def header(self):
        """ """
        return self._header
    
    @property
    def meta(self):
        """ """
        return self._meta

    @property
    def targetname(self):
        """ name of the target (from self.meta). 
        None if no meta or name not in meta"""
        if self.meta is None:
            return None
        
        return self.meta.get("ztfname")

    @property
    def obsdate(self):
        """ """
        if self.meta is None:
            return None
        
        if "date" not in self.meta.index or self.meta["date"] is None:
            warnings.warn("Unknown date for the given spectrum")
            return None

        from astropy.time import Time
        from datetime import datetime
        return Time(datetime.strptime(self.meta["date"], '%Y%m%d'))


    @property
    def snidresult(self):
        """ """
        return self._snidresult
    
    # Derived
    @property
    def lbda(self):
        """ """
        return np.asarray(self.data["lbda"].values, dtype="float")
    
    @property
    def flux(self):
        """ """
        return np.asarray(self.data["flux"].values, dtype="float")
    
    def has_error(self):
        """ """
        return self.has_variance()
    
    def has_variance(self):
        """ """
        return "variance" in self.data
    
    @property
    def variance(self):
        """ """
        if not self.has_error():
            return None
        
        return np.asarray(self.data["variance"].values, dtype="float")
        
    @property
    def error(self):
        """ """
        if not self.has_error():
            return None
        
        return np.sqrt(self.variance)

    @property
    def filename(self):
        """ """
        if not hasattr(self,"_filename"):
            return None
        return self._filename
    
# =============== #
#                 #
#                 #
#                 #
# =============== #
class Spectra( object ):
    """ Spectrum Collection """
    def __init__(self, spectra, use_dask=False):
        """ 
        spectra: [list]
            list of spectrum
        """
        self.set_spectra(spectra)
        self._use_dask = use_dask
        
    @classmethod
    def from_filenames(cls, filenames, use_dask=False, **kwargs):
        """ """
        spectra = [Spectrum.from_filename(file_, use_dask=use_dask, **kwargs) for file_ in filenames]
        return cls(spectra, use_dask=use_dask)
    
    @classmethod
    def from_directory(cls, directory, contains=None, startswith=None, extension=".ascii", use_dask=False, **kwargs):
        """ """
        from glob import glob
        glob_format = "*" if not startswith else f"{startswith}*"
        if contains is not None:
            glob_format += f"{contains}*"
        if extension is not None:
            glob_format +=f"{extension}"
        spectrafiles = glob(os.path.join(directory, glob_format))
        
        return cls.from_filenames(spectrafiles, use_dask=use_dask, **kwargs)

    # ================ #
    #    Method        #
    # ================ #
    def set_spectra(self, spectra):
        """ """
        self._spectra = spectra
        
    def get_meta(self, rebuild=False):
        """ """
        if rebuild:
            metas = self.call_down("meta", False)
            return pandas.DataFrame(metas).astype({"date":"<M8[ns]"})

        return self.meta

    def get_target_spectra(self, name):
        """ """
        return self.__class__( list(np.asarray(self.spectra)[np.asarray(self.meta.query(f"name == '{name}'").index)]),
                                  use_dask=self._use_dask )
    
    def get_spectra(self, sortby="date"):
        """ """
        if sortby is not None:
            index = self.meta.sort_values("date").index
            return np.asarray(self.spectra)[index]
        
        return self.spectra

    def get_redshift(self):
        """ """
        DEFAULT = [np.NaN, np.NaN]
        redshifts = np.asarray(self.call_down("get_redshift", True)).T
        # Clear nans out
        flagnan = np.any(np.isnan(redshifts), axis=0)
        if np.all(flagnan):
            return DEFAULT
        
        if np.any(flagnan):
            redshifts = redshifts.T[~flagnan].T

            
        weights = 1/redshifts[1]**2
        redshift = np.average(redshifts[0], weights=weights)
        dredshift = np.sqrt(1/np.sum(weights))
        return redshift, dredshift
        
    def show(self, ax=None, sortby="date", topfirst=True, legendprop={}, **kwargs):
        """ """
        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[7,4])
            ax = fig.add_axes([0.12,0.15,0.78,0.75])
        else:
            fig = ax.figure

        spectra = self.get_spectra(sortby)
        nspec = len(spectra)
        
        for i, spec in enumerate(spectra):
            _ = spec.show(ax=ax, offset=(nspec-i) if topfirst else i, 
                          normed=True, color=f"C{i}", 
                              legend=False, **{**dict(label="_meta_"),**kwargs})

        ax.legend(**{**dict(loc="best", frameon=False, fontsize="small"), **legendprop})
        return fig

    def show_target(self, name, ax=None, **kwargs):
        """ """
        spectarget = self.get_target_spectra(name)
        return spectarget.show(ax=ax, **kwargs)
    
    # -------- #
    # INTERNAL #
    # -------- #
    def map_down(self, what, margs, *args, **kwargs):
        """ """
        return [getattr(spec, what)(marg, *args, **kwargs)
                for spec, marg in zip(self.spectra, margs)]
    
    def call_down(self, what, isfunc, *args, **kwargs):
        """ """
        if isfunc:
            return [getattr(spec,what)(*args, **kwargs) for spec in self.spectra]
        return [getattr(spec,what) for spec in self.spectra]    
    
    # ================ #
    #   Properties     #
    # ================ #
    @property
    def meta(self):
        """ """
        if not hasattr(self,"_meta"):
            self._meta = self.get_meta(rebuild=True)
        return self._meta
    
    # Baseline    
    @property
    def spectra(self):
        """ """    
        return self._spectra

    @property
    def filenames(self):
        """ """
        return self.call_down("filename", isfunc=False)



import numpy as np

def mag_to_flux(mag, magerr=None, units="zp", zp=25.0, wavelength=None):
    """converts magnitude into flux
    Parameters
    ----------
    mag: [float or array]
        AB magnitude(s)
    magerr: [float or array] -optional-
        magnitude error if any
    units: [string] -optional-
        Unit system in which to return the flux:
        - 'zp':   units base on zero point as required for sncosmo fits
        - 'phys': physical units [erg/s/cm^2/A)
    zp: [float or array] -optional-
        zero point of for flux; required if units == 'zp'
    wavelength: [float or array] -optional-
        central wavelength of the photometric filter.
        In Angstrom; required if units == 'phys'
    Returns
    -------
    - float or array (if magerr is None)
    - float or array, float or array (if magerr provided)
    """
    if units not in ["zp", "phys"]:
        raise ValueError("units must be 'zp' or 'phys'")
    elif units == "zp":
        if zp is None:
            raise ValueError("zp must be float or array if units == 'zp'")
        flux = 10 ** (-(mag - zp) / 2.5)
    else:
        if wavelength is None:
            raise ValueError("wavelength must be float or array if units == 'phys'")
        flux = 10 ** (-(mag + 2.406) / 2.5) / wavelength ** 2

    if magerr is None:
        return flux

    dflux = np.abs(flux * (-magerr / 2.5 * np.log(10)))  # df/f = dcount/count
    return flux, dflux



""" Interact with the typingapp """

from . import io
import os
import pandas
import numpy as np


DB_PATH = os.path.join(io.IDR_PATH, "typingapp.db")

# typing app data
import sqlite3
con = sqlite3.connect(DB_PATH)


SORT_FAVORED = ["not ia","ib/c","ii","gal", "other",
                "sn ia","ia-norm",
                "ia-91bg", "ia-91t", "ia-other"]

def get_typing(typings_values, 
               sort_favored=SORT_FAVORED):
    """ """
    typings, values = typings_values
    typings[values==values.max()]
    if len(typings)  == 1:
        return typings[0]
    
    ltypings = list(typings)
    ltypings.sort(key = lambda i: sort_favored.index(i))
    return ltypings[0]



def parse_classification(line, 
                         min_review=2,
                         min_autotyping=3, 
                         min_generic_typing=4,
                         unclear_limit=0.5):
    """ """
    # not enough review, pass
    if line.nreviews <=min_review: # not enough
        classification = "None"
        
    # easy, all the same 
    elif line.ntypes == 1 and line.ntypings[0] >= min_autotyping: 
        classification =  line.typing[0]
        
    # Typing ok, Sub-tyiping not
    elif np.all(['ia' in k and "not ia" not in k for k in line.typing]) and line.nreviews >=min_generic_typing:
        classification = "ia"

    # all non-ia ?
    elif np.all([k in Classifications._NONIA_CASES for k in line.typing])  and line.nreviews >=min_generic_typing:
        classification = "nonia"
        

    # saving unclears if possible
    elif "unclear" in line["typing"]:
        info = dict(zip(line["typing"],line["ntypings"]))
        n_unclears = info["unclear"] 
        n_not = np.sum([v for k,v in info.items() if k !="unclear"])

        if n_unclears < (n_unclears+n_not)*unclear_limit and n_not >min_generic_typing:
            new_line = clear_unclear(line)
            classification = parse_classification(new_line)
        else:
            classification = "confusing"
    # Typing ok, Sub-tyiping not        
    else:
        classification = "confusing"
        
    return classification


def clear_unclear(line):
    """ remove the 'unclear' component from the given typing row """
    typing = list(line.typing)
    if "unclear" not in typing:
        return line # nothing to do
    
    ntypings = list(line.ntypings)
    index_unclear = typing.index("unclear")
    typing.pop(index_unclear)
    ntypings.pop(index_unclear)
    return pandas.Series({"typing": typing,
                          "ntypings": ntypings,
                          "ntypes": len(ntypings),
                          "nreviews": np.sum(ntypings)})


class _DBApp_():
    _DB_NAME = None

    def __init__(self):
        """ """
        self.load()
    
    @classmethod
    def data_from_db(cls, sql=None):
        """ """
        basequery = f"SELECT * FROM {cls._DB_NAME}"
        if sql is not None:
            basequery +=f" {sql}"
            
        return pandas.read_sql_query(basequery, con)
        
        
    def load(self):
        """ """
        self._data = self.data_from_db()
    
    # =============== #
    #   Properties    #
    # =============== #
    @property
    def data(self):
        """ """
        return self._data

# =================== #
#                     #
#  USERS              #
#                     #
# =================== #    
class _Users_( _DBApp_ ):
    _DB_NAME = "Users"
    
    def load(self):
        """ """
        self._data = self.data_from_db()[["id","username","name","email","date_added"]]

# =================== #
#                     #
#  CLASSIFICATIONS    #
#                     #
# =================== #
class Reports( _DBApp_ ):
    _DB_NAME = "Classifications"
    
    def load(self):
        """ """
        self._data = self.data_from_db(sql="where kind='report'")

# =================== #
#                     #
#  CLASSIFICATIONS    #
#                     #
# =================== #
class Classifications( _DBApp_ ):
    _DB_NAME = "Classifications"
    _NONIA_CASES = ["not ia", "ii","ib/c","gal","other"]

    MASTER_USER = ["jesper", "Joel Johansson", "Kate Maguire",
                   "Mathew Smith", "Umut","Georgios Dimitriadis"]

    def load(self):
        """ """
        self._data = self.data_from_db(sql="where kind='typing'")
        self._data["value"] = self._data["value"].str.strip().str.lower()

    def load_sample(self):
        """ """
        from . import get_sample
        self._sample = get_sample()

    def load_classification(self, min_review=2, min_autotyping=3, min_generic_typing=4,
                                  min_review_master=1, min_autotyping_master=2,
                                  min_generic_typing_master=42):
        """ """
        self._load_classification(min_review=min_review, min_autotyping=min_autotyping,
                                  min_generic_typing=min_generic_typing)
        # master
        master = self.get_masterclassification(False)
        master._load_classification(min_review=min_review_master, min_autotyping=min_autotyping_master,
                                  min_generic_typing=min_generic_typing_master)
        # master successful classifications
        mdata = master.data[~master.data["classification"].isin(["None","confusing"])
                            ].groupby("target_name").first()["classification"]
        mdata.name = "master_classification"        

        data = self.data.set_index("target_name")
        
        # merged
        datamerged = data.join(mdata).reset_index()
        datamerged["auto_classification"] = datamerged["classification"]
        mclassified = datamerged[~datamerged["master_classification"].isna()]["master_classification"]
        datamerged.loc[mclassified.index, "classification"] = mclassified
        
        self._data = datamerged
        
        
    def _load_classification(self, min_review=2, min_autotyping=3, min_generic_typing=4):
        # Normal
        class_df = self.get_classification_df()
        class_df["classification"] = class_df.apply(parse_classification, axis=1,
                                                    min_review=min_review, min_autotyping=min_autotyping,
                                                    min_generic_typing=min_generic_typing)

        # are you reloading ?
        if "classification" in self.data:
            _ = self.data.pop("classification")
            
        self._data = self.data.set_index("target_name").join(class_df["classification"]).reset_index()


        
    # -------- #
    # GETTER   #
    # -------- #
    def get_masterclassification(self, incl_unclear=True):
        """ """
        this = self.__class__()
        MASTER_ID = _Users_().data.set_index("name").loc[self.MASTER_USER]["id"].values
        this._data = self.data[self.data["user_id"].isin(MASTER_ID)]
        if not incl_unclear:
            this._data = this._data[~(this._data["value"] == "unclear")]
            
        return this

    def get_classification_stats(self, based_on="classification", **kwargs):
        """ Get the distribution of classifications made. 

        Parameters
        ----------
        based_on: str
            classification key to used (from self.data); could be:
            - classification
            - master_classification
            - auto_classification

        **kwargs goes to get_data()

        Returns
        -------
        pandas.Series
            groupby().size() serie.
        """
        if "classification" not in self.data:
            self.load_classification()

        data = self.get_data(**kwargs)
        return data.groupby("target_name").first().groupby(based_on).size()
    
    
    def get_data(self, incl_unclear=True, min_classification=None,
                     types=None, 
                     goodcoverage=None, redshift_range=None,
                     x1_range=None, c_range=None, **kwargs):
        """ 

        incl_unclear: [bool] -optional-
            should the unclear typing be included ?

        min_classification: [None or int] -optional-
            minimum number of typing for this target. 
            'unclear' are not counted as classifications
            (this uses self.get_ntyped()) 


        goodcoverage: [None or bool] -optional-
            # sample.get_data() options
            should we only use target with good lightcurve sampling ?
            - None: no selection
            - True: only good sampling targets
            - False: only bad sample targets
            
        redshift_range, x1_range, c_range: [None or [min, max] ] -optional-
            # sample.get_data() options
            range to limit redshift, x1 or c, respectively
        
        **kwargs goes to self.sample.get_data()

        Returns
        -------
        DataFrame
        """
        data = self.data.copy()
        if not incl_unclear:
            data = data[data["value"] != "unclear"]
            
        if min_classification is not None and min_classification>0:
            ntyped = self.get_ntyped(incl_unclear=False)
            data = data[data["target_name"].isin(ntyped[ntyped>=min_classification].index)]
            
        if types is not None:
            data = data[data["value"].isin(np.atleast_1d(types))]
            
        # Sample
        _locals = locals()
        prop_sample = kwargs
        for k in ["goodcoverage", "redshift_range", "x1_range", "c_range"]:
            if _locals[k] is not None:
                prop_sample[k] = _locals[k]
                
        if len(prop_sample) >0:
            names_to_keep = self.sample.get_data(**prop_sample).index
            data = data[data["target_name"].isin(names_to_keep)]
            
        return data

    def get_ntyped(self, incl_unclear=False):
        """ series of the n"""
        return self.get_data(incl_unclear=incl_unclear).groupby("target_name").size().sort_values()
    
    def get_target_data(self, name):
        """ """
        return self.data.query(f"target_name == '{name}'")
    
    def get_of_types(self, types):
        """ get entry associated to the given type (lower case) """
        return 
                         
    def get_nonia(self, incl_unclear=False):
        """ """
        list_to_get = self._NONIA_CASES.copy()
        if incl_unclear:
            list_to_get+="unclear"
            
        return self.data[self.data["value"].isin(list_to_get)]

    def get_classification_df(self):
        """ """
        data = self.get_data(incl_unclear=True)
        typings = data.groupby(["target_name"])["value"].apply(np.unique, return_counts=True)
        types =  pandas.DataFrame(typings.to_list(), columns=["typing", "ntypings"],
                                index=typings.index)#.explode(["typing","ntypings"])
        types["ntypes"] = types["ntypings"].apply(len)
        types["nreviews"] = types["ntypings"].apply(np.sum)
        return types
    
    # -------- #
    # PLOTTER  #
    # -------- #
    def show_classification_stats(self, ax=None, subtypes=None, redshift_range=None, 
                                  based_on="classification", explode=None, startangle=0,
                                  dataprop={}, **kwargs):
        """ """
        import matplotlib.pyplot as plt
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure


        class_stat = self.get_classification_stats(based_on, 
                                                  redshift_range=redshift_range, **dataprop)
        if subtypes is not None:
            class_stat = class_stat[[k for k in subtypes if k in class_stat]]

        if type(explode) == float:
            explode = np.full_like(class_stat.values, explode, dtype="float")
            
        prop = {**dict(explode=explode, autopct='%.0f%%', startangle=startangle), **kwargs}
        _ = ax.pie(class_stat.values, labels=class_stat.index, **prop)

        ax.set_title(f"{np.sum(class_stat.values)} Supernovae", color="tab:grey")
        return fig

    
    def show_classifications(self, ax=None, fig=None, 
                                 classifications=None, per_target=True):
        """ """
        if classifications is not None:
            ndatas = classifications.value_counts().sort_values()
        elif per_target:
            ndatas = self.get_classification().value_counts().sort_values()
        else:
            ndatas = self.data.groupby(["value"]).size().sort_values()
        
        if ax is None:
            if fig is None:
                import matplotlib.pyplot as mpl
                fig = mpl.figure(figsize=[7,4])
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure
        
        hbars = ax.barh(ndatas.index, ndatas, tick_label=ndatas.index)
        ax.tick_params("y", labelsize="small", color="C0", length=0)

        ax.bar_label(hbars, fmt='%d', color="C0", padding=1, 
                     fontsize="x-small")

        clearwhich = ["bottom","right","top", "left"] # "bottom"
        [ax.spines[which].set_visible(False) for which in clearwhich]
        ax.set_xticks([])
        return fig
        
    # ============== #
    #  Properties    #
    # ============== #
    @property
    def sample(self):
        """ """
        if not hasattr(self, "_sample"):
            self.load_sample()
            
        return self._sample

    
    @property
    def types(self):
        """ list of all possible types"""
        return self.data["value"].unique()

    @property
    def ia_types(self):
        """ list of all possible types"""
        alltypes = self.types
        return alltypes[["ia" in _ and _ !="not ia" for _ in alltypes ]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

"""
from distutils.core import setup
from setuptools import setup, find_packages


packages = find_packages()

VERSION = '0.7.1'
        
setup(name='ztfidr',
      version=VERSION,
      description='Tools for ZTF Ia *Internal*DataReleases',
      author='Mickael Rigault',
      author_email='m.rigault@ipnl.in2p3.fr',
      url='https://github.com/MickaelRigault/ztfdr',
      packages=packages,
#      package_data={'ztfdr': ['data/*']},
#      scripts=["bin/__.py",]
     )
# End of setupy.py ========================================================


